unit untValueArea;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, Menus, Buttons;

type
  TValueColor = record
    BarLeft,
    BarRight,
    BarGrid,
    BarAvg,
    HistoryGraph,
    HistoryGrid,
    HistoryAvg    : TColor;
  end;

  TValueOption = (voBar, voBarGrid, voBarAvg, voHistory, voHistoryGrid, voHistoryAvg, voHistoryLeft, voMinMax);

  TValueItem = record
    ID      : Integer;
    Caption : String;
    Hint    : String;
    VFormat : String;
    Value   : Real;
    Min     : Real;
    Max     : Real;
    Step    : Real;
    History : Array[0..24] of Real;
    Average : Real;
    Options : Set of TValueOption;
  end;

  TValueArea = class; // forward declaration

  TValueData = class(TComponent)
  private
    FColor   : TValueColor;
    FSection : String;
    FItems   : Array of TValueItem;
    FAreas   : Array of TValueArea;
  protected
    function  GetItem(ID : Integer): TValueItem;
    procedure SetItem(ID : Integer; const Value: TValueItem);
    procedure AddArea(Area: TValueArea);
    procedure DelArea(Area: TValueArea);
    function  ValidID(const ID: Integer): Boolean;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure SaveColors;
    procedure RestoreColors;
    procedure SaveOptions;
    procedure RestoreOptions;
    procedure Clear;
    procedure UpdateAreas;
    function  IndexToID(const Index: Integer): Integer;
    function  CountItems: Integer;
    procedure AddItem(const ID: Integer; const Caption, Hint, VFormat: String);
    procedure SetData(const ID: Integer; const RValue: Real);
    property  Item[ID : Integer] : TValueItem read GetItem write SetItem; default;
  published
    property  Color : TValueColor read FColor write FColor;
    property  IniSection : String read FSection write FSection;
  end;

  TValueArea = class(TCustomPanel)
  private
    FSection          : String;
    FData             : TValueData;
    FPopCoords        : TPoint;
    FRow              : Integer;
    FCell             : Integer;
    FPreview          : Boolean;
  protected
    procedure Resize; override;
    procedure GetRowCell;
    procedure PopupMenuPopup(Sender: TObject);
    procedure DeleteRow(Sender: TObject);
    procedure AddCell(Sender: TObject);
    procedure ChangeCell(Sender: TObject);
    procedure DeleteCell(Sender: TObject);
    procedure Options(Sender: TObject);
    procedure DoubleClick(Sender: TObject);
    procedure SetData(Value : TValueData);
    procedure AddRow(Sender: TObject);
    procedure AddCellAt(const Row: Integer; const ID: Integer);
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure SaveItems;
    procedure RestoreItems;
    procedure Update; override;
  published
    property Align;
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BorderStyle;
    property BorderWidth;
    property Data : TValueData read FData write SetData;
    property IniSection : String read FSection write FSection;
    property ParentShowHint;
    property ShowHint;
  end;

  TValueRow = class(TPanel)
  private
    FArea : TValueArea;
  protected
    procedure Resize; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure AddCell(const ID : Integer);
  end;

  TValueCell = class(TCustomControl)
  private
    FID   : Integer;
    FArea : TValueArea;
    procedure SetID(Value: Integer);
  protected
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property BorderWidth;
    property ID : Integer read FID write SetID;
  end;

  TfrmValueOptions = class(TForm)
    cldColor: TColorDialog;
    gbxItem: TGroupBox;
    cbxBar: TCheckBox;
    cbxBarAvg: TCheckBox;
    cbxBarGrid: TCheckBox;
    cbxHistory: TCheckBox;
    cbxHistoryAvg: TCheckBox;
    cbxHistoryGrid: TCheckBox;
    cbxHistoryLeft: TCheckBox;
    cbxMinMax: TCheckBox;
    lblMinimum: TLabel;
    edtMin: TEdit;
    lblMaximum: TLabel;
    edtMax: TEdit;
    lblStep: TLabel;
    edtStep: TEdit;
    gbxBar: TGroupBox;
    btnBarLeft: TButton;
    btnBarRight: TButton;
    btnBarAvg: TButton;
    btnBarGrid: TButton;
    gbxHistory: TGroupBox;
    btnHistoryGraph: TButton;
    btnHistoryAverage: TButton;
    btnHistoryGrid: TButton;
    gbxPreview: TGroupBox;
    Panel1: TPanel;
    Bevel1: TBevel;
    Panel2: TPanel;
    btbCancel: TBitBtn;
    btbOk: TBitBtn;
    pnlPreview: TPanel;
    procedure cbxBarClick(Sender: TObject);
    procedure cbxBarAvgClick(Sender: TObject);
    procedure cbxBarGridClick(Sender: TObject);
    procedure cbxHistoryClick(Sender: TObject);
    procedure cbxHistoryAvgClick(Sender: TObject);
    procedure cbxHistoryGridClick(Sender: TObject);
    procedure cbxHistoryLeftClick(Sender: TObject);
    procedure cbxMinMaxClick(Sender: TObject);
    procedure edtMinExit(Sender: TObject);
    procedure edtMaxExit(Sender: TObject);
    procedure edtStepExit(Sender: TObject);
    procedure btnBarLeftClick(Sender: TObject);
    procedure btnBarRightClick(Sender: TObject);
    procedure btnBarAvgClick(Sender: TObject);
    procedure btnBarGridClick(Sender: TObject);
    procedure btnHistoryGraphClick(Sender: TObject);
    procedure btnHistoryAverageClick(Sender: TObject);
    procedure btnHistoryGridClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
    FData : TValueData;
    procedure Preview;
  public
    { Public declarations }
    Item : TValueItem;
    procedure UpdateAppearance(oFont : TFont; oDialog, oWindow : TColor);
  end;

procedure Register;

implementation

{$R *.DFM}

uses
  IniFiles;

function DummyItem : TValueItem;
var
  Item : TValueItem;
begin
  FillChar(Item, SizeOf(TValueItem), 0);
  with Item do
    begin
      Caption := 'Dummy';
      Max     := 200;
      Step    := 25;
      Options := [voBar, voBarGrid, voBarAvg, voHistory, voHistoryGrid, voHistoryAvg, voMinMax];
    end;
  Result := Item;
end;

procedure TValueData.UpdateAreas;
var
  i : Integer;
begin
  if (Length(FAreas) > 0) then
    for i := 0 to High(FAreas) do
      if Assigned(FAreas[i]) then
        FAreas[i].Update;
end;

procedure TValueData.Clear;
var
  i, iH : Integer;
begin
  for i := 0 to High(FItems) do
    with FItems[i] do
      begin
        Value   := 0;
        Average := 0;
        for iH := 0 to High(History) do
          History[iH] := -1;
      end;
  UpdateAreas;
end;

procedure TValueData.SaveColors;
var
  sFile, sColors : String;
begin
  with FColor do
    sColors := Format('%d %d %d %d %d %d %d', [BarLeft, BarRight, BarGrid, BarAvg, HistoryGraph, HistoryGrid, HistoryAvg]);
  sFile := ChangeFileExt(ParamStr(0), '.ini');
  with TIniFile.Create(sFile) do
    begin
      WriteString(FSection, 'Colors', sColors);
      UpdateFile;
      Free;
    end;
end;

procedure TValueData.RestoreColors;
var
  sFile, sColors, S, C : String;
  i : Integer;
begin
  sFile := ChangeFileExt(ParamStr(0), '.ini');
  with TIniFile.Create(sFile) do
    begin
      sColors := ReadString(FSection, 'Colors', '');
      Free;
    end;
  S := sColors + ' ';
  try
    i := Pos(' ', S);
    if (i = 0) then Exit;
    C := Copy(S, 1, i - 1);
    FColor.BarLeft := StrToInt(C);
    S := Copy(S, i + 1, MAXINT);

    i := Pos(' ', S);
    if (i = 0) then Exit;
    C := Copy(S, 1, i - 1);
    FColor.BarRight := StrToInt(C);
    S := Copy(S, i + 1, MAXINT);

    i := Pos(' ', S);
    if (i = 0) then Exit;
    C := Copy(S, 1, i - 1);
    FColor.BarGrid := StrToInt(C);
    S := Copy(S, i + 1, MAXINT);

    i := Pos(' ', S);
    if (i = 0) then Exit;
    C := Copy(S, 1, i - 1);
    FColor.BarAvg := StrToInt(C);
    S := Copy(S, i + 1, MAXINT);

    i := Pos(' ', S);
    if (i = 0) then Exit;
    C := Copy(S, 1, i - 1);
    FColor.HistoryGraph := StrToInt(C);
    S := Copy(S, i + 1, MAXINT);

    i := Pos(' ', S);
    if (i = 0) then Exit;
    C := Copy(S, 1, i - 1);
    FColor.HistoryGrid := StrToInt(C);
    S := Copy(S, i + 1, MAXINT);

    i := Pos(' ', S);
    if (i = 0) then Exit;
    C := Copy(S, 1, i - 1);
    FColor.HistoryAvg := StrToInt(C);
    S := Copy(S, i + 1, MAXINT);
  except
    //
  end;
end;

constructor TValueData.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FSection := 'ValueData';
  SetLength(FItems, 0);
  SetLength(FAreas, 0);
  with FColor do
    begin
      BarLeft      := clYellow;
      BarRight     := clAqua;
      BarGrid      := clBtnFace;
      BarAvg       := clGray;
      HistoryGraph := clRed;
      HistoryGrid  := clGray;
      HistoryAvg   := clBlue;
    end;
end;

destructor TValueData.Destroy;
var
  i : Integer;
begin
  if (Length(FAreas) > 0) then
    for i := 0 to High(FAreas) do
      if Assigned(FAreas[i]) then
        FAreas[i].SetData(Nil);
  inherited Destroy;
end;

procedure TValueData.SaveOptions;
var
  sFile, sOptions : String;
  iItem : Integer;
begin
  sFile := ChangeFileExt(ParamStr(0), '.ini');
  with TIniFile.Create(sFile) do
    begin
      for iItem := 1 to CountItems do
        with GetItem(IndexToID(iItem)) do
          begin
            sOptions := Trim(Format(VFormat, [Min])) + ' ' +
                        Trim(Format(VFormat, [Max])) + ' ' +
                        Trim(Format(VFormat, [Step])) + ' ' +
                        Trim(Format('%d%d%d%d%d%d%d%d', [Ord(voBar in Options),
                                                         Ord(voBarGrid in Options),
                                                         Ord(voBarAvg in Options),
                                                         Ord(voHistory in Options),
                                                         Ord(voHistoryGrid in Options),
                                                         Ord(voHistoryAvg in Options),
                                                         Ord(voHistoryLeft in Options),
                                                         Ord(voMinMax in Options)]));
            WriteString(FSection, 'Options' + IntToSTr(ID), sOptions);
          end;
      UpdateFile;
      Free;
    end;
end;

procedure TValueData.RestoreOptions;
var
  sIniFile, S, T : String;
  iItem, ID, i : Integer;
  Item : TValueItem;
begin
  sIniFile := ChangeFileExt(ParamStr(0),'.ini');
  with TIniFile.Create(sIniFile) do
    begin
      for iItem := 1 to CountItems do
        begin
          ID := IndexToID(iItem);
          S   := ReadString(FSection, Format('Options%d', [ID]) , '');
          for i := 1 to Length(S) do
            if (S[i] = '.') or (S[i] = ',') then
              S[i] := DecimalSeparator;

          Item := GetItem(ID);
          Item.Options := [];
          i := Pos(' ', S);
          T := Trim(Copy(S, 1, i - 1));
          if (T='') then
            T := '0';
          Item.Min := StrToFloat(T);
          S := Copy(S, i + 1, MAXINT);
          i := Pos(' ', S);
          T := Trim(Copy(S, 1, i - 1));
          if (T='') then
            T := '0';
          Item.Max := StrToFloat(T);
          S := Copy(S, i + 1, MAXINT);
          i := Pos(' ', S);
          T := Trim(Copy(S, 1, i - 1));
          if (T='') then
            T := '0';
          Item.Step := StrToFloat(T);
          S := Trim(Copy(S, i + 1, MAXINT));
          if (Copy(S, 1, 1) = '1') then Item.Options := Item.Options + [voBar];
          if (Copy(S, 2, 1) = '1') then Item.Options := Item.Options + [voBarGrid];
          if (Copy(S, 3, 1) = '1') then Item.Options := Item.Options + [voBarAvg];
          if (Copy(S, 4, 1) = '1') then Item.Options := Item.Options + [voHistory];
          if (Copy(S, 5, 1) = '1') then Item.Options := Item.Options + [voHistoryGrid];
          if (Copy(S, 6, 1) = '1') then Item.Options := Item.Options + [voHistoryAvg];
          if (Copy(S, 7, 1) = '1') then Item.Options := Item.Options + [voHistoryLeft];
          if (Copy(S, 8, 1) = '1') then Item.Options := Item.Options + [voMinMax];

          SetItem(ID, Item);
        end;
      Free;
    end;
end;

//-------------------------------------------------------------------

procedure TValueCell.Paint;
var
  H, B, W, S, X, Y, i, S1, S2: Integer;
  R1, R2, R3: TRect;
  Flags: Longint;
  E :  TSize;
  rRange, rPos, D : Real;
  N : TFontName;
  V : String;
  Item : TValueItem;
  oBit : TBitmap;
begin
  with Canvas do
    if (FID < 0) then
      begin
        Brush.Color := Color;
        FillRect(ClientRect);
        Exit;
      end;

  if FArea.FPreview then
    with TGroupBox(FArea.Owner) do
      Item := TfrmValueOptions(Owner).Item
  else
    if Assigned(FArea.FData) then
      Item := FArea.FData.GetItem(FID)
    else
      with Canvas do
        begin
          Brush.Color := Color;
          FillRect(ClientRect);
          FID := -1;
          Exit;
        end;

  oBit := TBitmap.Create;
  with oBit do
    begin
      Width  := ClientRect.Right;
      Height := ClientRect.Bottom;
      Canvas.Brush.Color := Color;
      Canvas.Brush.Style := bsSolid;
      Canvas.FillRect(ClientRect);
    end;

  with oBit.Canvas, Item do
    begin
      // Frame
      Font := Self.Font;
      H := TextHeight('0');
      B := BorderWidth * 2;
      R1 := Rect(0, H div 2 - 1, Width - B, Height - B);
      if Ctl3D then
        begin
          Inc(R1.Left);
          Inc(R1.Top);
          Brush.Color := clBtnHighlight;
          FrameRect(R1);
          OffsetRect(R1, -1, -1);
          Brush.Color := clBtnShadow;
        end
      else
        Brush.Color := clWindowFrame;
      FrameRect(R1);

      // Interior
      R2 := R1;
      InflateRect(R2, -3, - (H div 2));
      OffsetRect(R2, 0, 3);
      Inc(R2.Left);
      Brush.Color := Color;
      FillRect(R2);
    end;

  with Item do
    rRange := Max - Min;

  with oBit.Canvas, Item do
    if (rRange > 0) and (voBar in Options) then
      begin
        // Bar
        W           := R2.Right - R2.Left;
        rPos        := (Value-Min) / rRange;
        if (rPos < 0) then rPos := 0;
        if (rPos > 1) then rPos := 1;
        R3          := R2;
        R3.Right    := R3.Left + Round(W * rPos);
        Brush.Color := FArea.FData.FColor.BarLeft;
        FillRect(R3);
        R3          := R2;
        R3.Left     := R3.Left + Round(W * rPos);
        Brush.Color := FArea.FData.FColor.BarRight;
        FillRect(R3);

        // Gridlines
        if (Step = 0) then
          S := 0
        else
          S := Round(rRange/Step);
        if (voBarGrid in Options) and (Step > 0) and (S < (R2.Right - R2.Left - 2) / 2) then
          begin
            Pen.Color := FArea.FData.FColor.BarGrid;
            Pen.Width := 1;
            for i := 0 to S do
              begin
                rPos := (i * Step) / rRange;
                X := R2.Left + Round(rPos * (R2.Right - R2.Left - 2));
                Y := R2.Top;
                MoveTo(X, Y);
                Y := R2.Bottom;
                LineTo(X, Y);
              end;
          end;

        // Average
        if (voBarAvg in Options) then
          begin
            rPos        := (Average-Min) / rRange;
            if (rPos < 0) then rPos := 0;
            if (rPos > 1) then rPos := 1;
            Pen.Color := FArea.FData.FColor.BarAvg;
            Pen.Width := 3;
            X := R2.Left + Round(W * rPos);
            if (rPos = 0) then Inc(X, Pen.Width div 2);
            if (rPos = 1) then Dec(X, Pen.Width div 2);
            Y := R2.Top + 1;
            MoveTo(X, Y);
            Y := R2.Bottom - 2;
            LineTo(X, Y);
          end;
      end;

  R3 := R2;
  with oBit.Canvas, Item do
    if (voHistoryLeft in Options) then
      begin
        S := Font.Size;
        N := Font.Name;
        if (Max > Value) then
          V := Format(VFormat, [Max])
        else
          V := Format(VFormat, [Value]);
        Font.Name := 'Arial';
        Font.Size := 100;
        E := TextExtent(V);
        S1 := Round(140 * (R2.Bottom - R2.Top - 8) / E.cy);
        S2 := Round(100 * (R2.Right - R2.Left - 8) / E.cx);
        if (S1 < S2) then
          Font.Size := S1
        else
          Font.Size := S2;
        W := TextWidth(V);
        R3.Right := R2.Right - W - 4;
        Font.Size := S;
        Font.Name := N;
      end;

  with oBit.Canvas, Item do
    if (rRange > 0) and (voHistory in Options) then
      begin
        // History
        // Gridlines
        if (Step = 0) then
          S := 0
        else
          S := Round(rRange/Step);
        if (voHistoryGrid in Options) and (Step > 0) and (S < (R3.Bottom - R3.Top - 4) / 2) then
          begin
            Pen.Color := FArea.FData.FColor.HistoryGrid;
            Pen.Width := 1;
            for i := 0 to S do
              begin
                rPos := (i * Step) / rRange;
                X := R3.Left + 2;
                Y := R3.Bottom - 2 - Round(rPos * (R3.Bottom - R3.Top - 4));
                MoveTo(X, Y);
                X := R3.Right - 2;
                LineTo(X, Y);
              end;
          end;

        // Average
        if (voHistoryAvg in Options) then
          begin
            Pen.Color := FArea.FData.FColor.HistoryAvg;
            Pen.Width := 2;
            rPos := (Average-Min) / rRange;
            if (rPos < 0) then rPos := 0;
            if (rPos > 1) then rPos := 1;
            X := R3.Left + 2;
            Y := R3.Bottom - 2 - Round(rPos * (R3.Bottom - R3.Top - 4));
            MoveTo(X, Y);
            X := R3.Right - 2;
            LineTo(X, Y);
          end;

        // Graph
        Pen.Color := FArea.FData.FColor.HistoryGraph;
        Pen.Width := 2;
        D := (R3.Right - R3.Left - 4) / High(History);
        for i := 0 to High(History) do
          begin
            rPos := (History[i]-Min) / rRange;
            if (rPos < 0) then rPos := 0;
            if (rPos > 1) then rPos := 1;
            X := R3.Left + 2 + Round(D * i);
            Y := R3.Bottom - 2 - Round(rPos * (R3.Bottom - R3.Top - 4));
            if (i = 0) then
              MoveTo(X, Y)
            else
              LineTo(X, Y);
          end;
      end;

  with oBit.Canvas, Item do
    if (rRange > 0) and (voMinMax in Options) then
      begin
        // Min - Max
        Brush.Style := bsClear;
        V := Format(VFormat, [Max]);
        E := TextExtent(V);
        TextRect(R2, 4, R2.Top + 2, V);

        if (((R2.Bottom - R2.Top) / E.cy) > 3) then
          begin
            V := Format(VFormat, [Average]);
            TextRect(R2, 4 + E.cx - TextWidth(V), R2.Top + (R2.Bottom - R2.Top - H) div 2, V);
          end;

        if (((R2.Bottom - R2.Top) / E.cy) > 2) then
          begin
            V := Format(VFormat, [Min]);
            TextRect(R2, 4 + E.cx - TextWidth(V), R2.Bottom - 2 - H, V);
          end;
        Brush.Style := bsSolid;
      end;

  with oBit.Canvas, Item do
    begin
      // Value
      S := Font.Size;
      N := Font.Name;
      Brush.Style := bsClear;
      if (Max > Value) then
        V := Format(VFormat, [Max])
      else
        V := Format(VFormat, [Value]);
      Font.Name := 'Arial';
      Font.Size := 100;
      E := TextExtent(V);
      S1 := Round(140 * (R2.Bottom - R2.Top - 8) / E.cy);
      S2 := Round(100 * (R2.Right - R2.Left - 8) / E.cx);
      if (S1 < S2) then
        Font.Size := S1
      else
        Font.Size := S2;
      V := Format(VFormat, [Value]);
      W := TextWidth(V);
      TextRect(R2, R2.Right - W - 4, R2.Top + ((R2.Bottom - R2.Top - 8) - TextHeight(V) ) div 2, V);
      Font.Size := S;
      Font.Name := N;
      Brush.Style := bsSolid;
    end;

  with oBit.Canvas, Item do
    if (Caption <> '') then
      begin
        // Caption
        R1 := Rect(8, 0, 0, H);
        Flags := DrawTextBiDiModeFlags(DT_SINGLELINE);
        DrawText(Handle, PChar(Caption), Length(Caption), R1, Flags or DT_CALCRECT);
        Brush.Color := Color;
        DrawText(Handle, PChar(Caption), Length(Caption), R1, Flags);
      end;

  Canvas.CopyRect(ClientRect, oBit.Canvas, ClientRect);
  oBit.Free;
end;

procedure TValueCell.SetID(Value: Integer);
begin
  if (Value <> FID) and Assigned(FArea.FData) then
    begin
      FID  := Value;
      Hint := FArea.FData.GetItem(FID).Hint;
    end;
end;

constructor TValueCell.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Align       := alClient;
  BorderWidth := 4;
  Parent      := TWinControl(AOwner);
  ParentColor := True;
end;

//-------------------------------------------------------------------

procedure TValueRow.Resize;
var
  i : Integer;
begin
  for i := 0 to ControlCount - 1 do
    TValueCell(Controls[i]).Width := Width div ControlCount;
end;

constructor TValueRow.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Align       := alclient;
  BevelOuter  := bvNone;
  Parent      := TWinControl(AOwner);
  ParentColor := True;
  FArea       := TValueArea(AOwner);
  Caption     := 'Right click to add a cell';
end;

procedure TValueRow.AddCell(const ID: Integer);
var
  i : Integer;
  C : TValueCell;
begin
  for i := 0 to ControlCount - 1 do
    TValueCell(Controls[i]).Align := alLeft;
  C := TValueCell.Create(Self);
  C.FArea      := FArea;
  C.OnDblClick := FArea.DoubleClick;
  C.ID         := ID;
  Resize;
end;

//-------------------------------------------------------------------

procedure TValueData.AddArea(Area: TValueArea);
var
  i, iAreas : Integer;
begin
  for i := 0 to High(FAreas) do
    if Assigned(FAreas[i]) then
      if (FAreas[i] = Area) then
        Exit;
  for i := 0 to High(FAreas) do
    if not(Assigned(FAreas[i])) then
      begin
        FAreas[i] := Area;
        Exit;
      end;
  iAreas := Length(FAreas);
  SetLength(FAreas, iAreas + 1);
  FAreas[iAreas] := Area;
end;

procedure TValueData.DelArea(Area: TValueArea);
var
  i : Integer;
begin
  if (Length(FAreas) > 0) then
    for i := 0 to High(FAreas) do
      if (FAreas[i] = Area) then
        FAreas[i] := Nil;
end;

function TValueData.IndexToID(const Index: Integer): Integer;
begin
  if (Index >= 1) and (Index <= Length(FItems)) then
    Result := FItems[Index-1].ID
  else
    Result := -1;
end;

function TValueData.ValidID(const ID: Integer): Boolean;
var
  i : Integer;
begin
  Result := False;
  for i := 0 to High(FItems) do
    if (FItems[i].ID = ID) then
      begin
        Result := True;
        Break;
      end;
end;

function  TValueData.CountItems: Integer;
begin
  Result := Length(FItems);
end;

function TValueData.GetItem(ID: Integer): TValueItem;
var
  i : Integer;
begin
  FillChar(Result, SizeOf(TValueItem), 0);
  Result.ID := -1;
  if ValidID(ID) then
    for i := 0 to High(FItems) do
      if (FItems[i].ID = ID) then
        begin
          Result := FItems[i];
          Break;
        end;
end;

procedure TValueData.SetItem(ID: Integer; const Value : TValueItem);
var
  i : Integer;
begin
  if ValidID(ID) then
    for i := 0 to High(FItems) do
      if (FItems[i].ID = ID) then
        begin
          FItems[i] := Value;
          Break;
        end;
end;

procedure TValueData.AddItem(const ID: Integer; const Caption, Hint, VFormat: String);
var
  i : Integer;
  Item : TValueItem;
begin
  if not(ValidID(ID)) then
    begin
      i := Length(FItems);
      SetLength(FItems, i + 1);
      FItems[i].ID := ID;
    end;
  Item := DummyItem;
  Item.ID      := ID;
  Item.Caption := Caption;
  Item.Hint    := Hint;
  Item.VFormat := VFormat;
  SetItem(ID, Item);
end;

procedure TValueData.SetData(const ID: Integer; const RValue: Real);
var
  Item : TValueItem;
  iH, i : Integer;
  R : Real;
begin
  if not(ValidID(ID)) then
    Exit;
  Item := GetItem(ID);
  with Item do
    begin
      i := 0;
      R := 0;
      Value := RValue;
      for iH := 0 to High(History) - 1 do
        History[iH] := History[iH+1];
      History[High(History)] := Value;
      for iH := 0 to High(History) do
        if (History[iH] > 0) then
          begin
            Inc(i);
            R := R + History[iH];
          end;
      if (i = 0) then
        Average := Value
      else
        Average := R / i;
    end;
  SetItem(ID, Item);
  UpdateAreas;
end;

//-------------------------------------------------------------------

procedure TValueArea.Resize;
var
  i : Integer;
begin
  for i := 0 to ControlCount - 1 do
    TValueRow(Controls[i]).Height := Height div ControlCount;
end;

procedure TValueArea.GetRowCell;
var
  i : Integer;
  P : TPoint;
begin
  FPopCoords := Mouse.CursorPos;
  FRow  := -1;
  FCell := -1;
  for i := 0 to ControlCount - 1 do
    with TValueRow(Controls[i]) do
      begin
        P := ScreenToClient(FPopCoords);
        if (P.Y > 0) and (P.Y < Height) then
          begin
            FRow := i;
            Break;
          end;
      end;
  if (FRow >= 0) then
    with TValueRow(Controls[FRow]) do
      for i := 0 to ControlCount - 1 do
        with TValueCell(Controls[i]) do
          begin
            P := ScreenToClient(FPopCoords);
            if (P.X > 0) and (P.X < Width) then
              begin
                FCell := i;
                Break;
              end;
          end;
end;

procedure TValueArea.PopupMenuPopup(Sender: TObject);
var
  i, ID : Integer;
  M : TMenuItem;
begin
  GetRowCell;
  if (FRow >= 0) and (FCell >=0) then
    ID := TValueCell(TValueRow(Controls[FRow]).Controls[FCell]).ID
  else
    ID := -2;
  with PopupMenu do
    begin
      Items[1].Enabled := (FRow >= 0);
      Items[3].Enabled := (FRow >= 0);
      Items[4].Enabled := (FCell >= 0);
      Items[5].Enabled := (FCell >= 0);
      if (FRow >= 0) and (FCell >=0) then
        Items[7].Enabled := (ID >= 0)
      else
        Items[7].Enabled := False;
    end;

  if not(Assigned(FData)) then
    Exit;

  if (PopupMenu.Items[3].Count = 2) then
    for i:= 0 to High(FData.FItems) do
      with FData.FItems[i] do
        begin
          M := TMenuItem.Create(PopupMenu.Items[3]);
          M.Caption := Caption;
          M.Hint    := Hint;
          M.Tag     := ID;
          M.OnClick := AddCell;
          PopupMenu.Items[3].Add(M);

          M := TMenuItem.Create(PopupMenu.Items[4]);
          M.Caption := Caption;
          M.Hint    := Hint;
          M.Tag     := ID;
          M.OnClick := ChangeCell;
          PopupMenu.Items[4].Add(M);
        end;

  for i:= 0 to PopupMenu.Items[4].Count - 1 do
    with PopupMenu.Items[4].Items[i] do
      Checked := (ID = Tag);
end;

procedure TValueArea.SetData(Value : TValueData);
begin
  if (Value <> FData) then
    begin
      if Assigned(FData) then
        FData.DelArea(Self);
      if Assigned(Value) then
        Value.AddArea(Self);
      FData := Value;
      Update;
    end;
end;

constructor TValueArea.Create(AOwner: TComponent);
var
  M, MSub : TMenuItem;
begin
  inherited Create(AOwner);
  FSection       := 'ValueArea';
  BevelOuter     := bvNone;
  Parent         := TWinControl(AOwner);
  ParentColor    := True;
  TabStop        := False;
  FPreview       := False;
  ParentShowHint := True;
  OnDblClick     := DoubleClick;

  if (csDesigning in ComponentState) then
    Exit // Don't build popupmenu in design mode
  else
    Caption   := 'Right click to add a row';

  PopupMenu   := TPopupMenu.Create(Self);
  PopupMenu.OnPopup := PopupMenuPopup;

  M := TMenuItem.Create(PopupMenu);
  M.Caption := '&Add row';
  M.Hint    := 'Add a row to this area';
  M.OnClick := AddRow;
  PopupMenu.Items.Add(M);

  M := TMenuItem.Create(PopupMenu);
  M.Caption := '&Delete row';
  M.Hint    := 'Delete this row';
  M.OnClick := DeleteRow;
  PopupMenu.Items.Add(M);

  M := TMenuItem.Create(PopupMenu);
  M.Caption := '-';
  PopupMenu.Items.Add(M);

  M := TMenuItem.Create(PopupMenu);
  M.Caption := 'Add &cell';
  M.Hint    := 'Add a cell to this row';
  PopupMenu.Items.Add(M);

  MSub := TMenuItem.Create(M);
  MSub.Caption := 'Empty';
  MSub.Hint    := 'Empty cell';
  MSub.Tag     := -1;
  MSub.OnClick := AddCell;
  M.Add(MSub);

  MSub := TMenuItem.Create(M);
  MSub.Caption := '-';
  MSub.Tag     := -2;
  M.Add(MSub);

  M := TMenuItem.Create(PopupMenu);
  M.Caption := 'C&hange cell';
  M.Hint    := 'Change the contents of this cell';
  PopupMenu.Items.Add(M);

  MSub := TMenuItem.Create(M);
  MSub.Caption := 'Empty';
  MSub.Hint    := 'Empty cell';
  MSub.Tag     := -1;
  MSub.OnClick := ChangeCell;
  M.Add(MSub);

  MSub := TMenuItem.Create(M);
  MSub.Caption := '-';
  MSub.Tag     := -2;
  M.Add(MSub);

  M := TMenuItem.Create(PopupMenu);
  M.Caption := 'D&elete cell';
  M.Hint    := 'Delete this cell';
  M.OnClick := DeleteCell;
  PopupMenu.Items.Add(M);

  M := TMenuItem.Create(PopupMenu);
  M.Caption := '-';
  PopupMenu.Items.Add(M);

  M := TMenuItem.Create(PopupMenu);
  M.Caption := '&Options';
  M.Hint    := 'Set item options';
  M.OnClick := Options;
  PopupMenu.Items.Add(M);
end;

destructor TValueArea.Destroy;
begin
  PopupMenu.Free;
  if Assigned(FData) then
    FData.DelArea(Self);
  inherited Destroy;
end;


procedure TValueArea.SaveItems;
var
  sIniFile : String;
  sItems : string;
  iR, iC : integer;
begin
  sIniFile := ChangeFileExt(ParamStr(0),'.ini');
  with TIniFile.Create(sIniFile) do
    begin
      sItems := '';
      for iR := 0 to ControlCount - 1 do
        with TValueRow(Controls[iR]) do
          for iC := 0 to ControlCount - 1 do
            with TValueCell(Controls[iC]) do
              sItems := format('%s %d.%d', [sItems, iR, ID]);
      WriteString(FSection, 'Items', Trim(sItems));
      UpdateFile;
      Free;
    end;
end;

procedure TValueArea.RestoreItems;
var
  sIniFile, sItems : String;
  i, j, Row, ID : Integer;
begin
  sIniFile := ChangeFileExt(ParamStr(0),'.ini');
  with TIniFile.Create(sIniFile) do
    begin
      sItems := ReadString(FSection, 'Items', '') + ' ';
      i := Pos(' ', sItems);
      while (i > 1) do
        begin
          j := Pos('.', sItems);
          Row := StrToInt(Copy(sItems, 1, j - 1));
          ID := StrToInt(Copy(sItems, j + 1, i - 1 - j));
          AddCellAt(Row, ID);
          sItems := Copy(sItems, i + 1, MAXINT);
          i := Pos(' ', sItems);
        end;
      Free;
    end;
end;

procedure TValueArea.AddRow(Sender: TObject);
var
  i : Integer;
begin
  for i := 0 to ControlCount - 1 do
    TValueRow(Controls[i]).Align := alTop;
  with TValueRow.Create(Self) do
    FArea := Self;
  Resize;
end;

procedure TValueArea.DeleteRow(Sender: TObject);
var
  i : Integer;
begin
  if (FRow < 0) then
    Exit;
  TValueRow(Controls[FRow]).Free;
  for i := 0 to ControlCount - 1 do
    if (i < (ControlCount - 1)) then
      TValueRow(Controls[i]).Align := alTop
    else
      TValueRow(Controls[i]).Align := alClient;
  Resize;
end;

procedure TValueArea.AddCell(Sender: TObject);
var
  ID : Integer;
begin
  if (FRow < 0) or not(Assigned(FData)) then
    Exit;
  ID := -1;
  if (Sender is TMenuItem) then
    ID := TMenuItem(Sender).Tag;
  with TValueRow(Controls[FRow]) do
    if FData.ValidID(ID) then
      AddCell(ID)
    else
      AddCell(-1);
  Update;
end;

procedure TValueArea.ChangeCell(Sender: TObject);
var
  ID : Integer;
begin
  if (FRow < 0) or not(Assigned(FData)) then
    Exit;
  ID := -1;
  if (Sender is TMenuItem) then
    ID := TMenuItem(Sender).Tag;
  with TValueRow(Controls[FRow]) do
    if FData.ValidID(ID) then
      TValueCell(Controls[FCell]).ID := ID
    else
      TValueCell(Controls[FCell]).ID := -1;
  Update;
end;

procedure TValueArea.DeleteCell(Sender: TObject);
var
  i : Integer;
begin
  if (FRow < 0) then
    Exit;
  if (FCell < 0) then
    Exit;
  with TValueRow(Controls[FRow]) do
    begin
      TValueCell(Controls[FCell]).Free;
      for i := 0 to ControlCount - 1 do
        if (i < (ControlCount - 1)) then
          TValueCell(Controls[i]).Align := alLeft
        else
          TValueCell(Controls[i]).Align := alClient;
      Resize;
    end;
end;

procedure TValueArea.AddCellAt(const Row: Integer; const ID: Integer);
var
  i : Integer;
begin
  if (Row < 0) or not(Assigned(FData)) then
    Exit;
  for i := ControlCount to Row do
    AddRow(Self);
  with TValueRow(Controls[Row]) do
    if FData.ValidID(ID) then
      AddCell(ID)
    else
      AddCell(-1);
  Update;
end;

procedure TValueArea.Options(Sender: TObject);
var
  ID : Integer;
begin
  if (FRow < 0) or not(Assigned(FData)) then
    Exit;
  if (FCell < 0) then
    Exit;
  with TValueRow(Controls[FRow]) do
    ID := TValueCell(Controls[FCell]).ID;
  if not(FData.ValidID(ID)) then
    Exit;
  with TfrmValueOptions.Create(Self) do
    begin
      UpdateAppearance(Self.Font, Self.Color, clWhite);
      Item := FData.GetItem(ID);
      if (ShowModal = mrOk) then
        FData.SetItem(ID, Item);
      Free;
    end;
  Update;
  FData.UpdateAreas;
end;

procedure TValueArea.DoubleClick(Sender: TObject);
begin
  if FPreview then
    Exit;
  GetRowCell;
  Options(Sender);
end;

procedure TValueArea.Update;
var
  iC, iR : Integer;
begin
  inherited Update;
  for iR := 0 to ControlCount - 1 do
    with TValueRow(Controls[iR]) do
      for iC := 0 to ControlCount - 1 do
        with TValueCell(Controls[iC]) do
          Paint; 
end;


//-------------------------------------------------------------------

procedure TfrmValueOptions.cbxBarClick(Sender: TObject);
begin
  with Item do
    if cbxBar.Checked then
      Options := Options + [voBar]
    else
      Options := Options - [voBar];
  cbxBarGrid.Enabled := cbxBar.Checked;
  cbxBarAvg.Enabled := cbxBar.Checked;
  Preview;
end;

procedure TfrmValueOptions.cbxBarAvgClick(Sender: TObject);
begin
  with Item do
    if cbxBarAvg.Checked then
      Options := Options + [voBarAvg]
    else
      Options := Options - [voBarAvg];
  Preview;
end;

procedure TfrmValueOptions.cbxBarGridClick(Sender: TObject);
begin
  with Item do
    if cbxBarGrid.Checked then
      Options := Options + [voBarGrid]
    else
      Options := Options - [voBarGrid];
  Preview;
end;

{ - - - - - - - - - - - }

procedure TfrmValueOptions.cbxHistoryClick(Sender: TObject);
begin
  with Item do
    if cbxHistory.Checked then
      Options := Options + [voHistory]
    else
      Options := Options - [voHistory];
  cbxHistoryGrid.Enabled := cbxHistory.Checked;
  cbxHistoryAvg.Enabled  := cbxHistory.Checked;
  cbxHistoryLeft.Enabled := cbxHistory.Checked;
  Preview;
end;

procedure TfrmValueOptions.cbxHistoryAvgClick(Sender: TObject);
begin
  with Item do
    if cbxHistoryAvg.Checked then
      Options := Options + [voHistoryAvg]
    else
      Options := Options - [voHistoryAvg];
  Preview;
end;

procedure TfrmValueOptions.cbxHistoryGridClick(Sender: TObject);
begin
  with Item do
    if cbxHistoryGrid.Checked then
      Options := Options + [voHistoryGrid]
    else
      Options := Options - [voHistoryGrid];
  Preview;
end;

procedure TfrmValueOptions.cbxHistoryLeftClick(Sender: TObject);
begin
  with Item do
    if cbxHistoryLeft.Checked then
      Options := Options + [voHistoryLeft]
    else
      Options := Options - [voHistoryLeft];
  Preview;
end;

{ - - - - - - - - - - - }

procedure TfrmValueOptions.cbxMinMaxClick(Sender: TObject);
begin
  with Item do
    if cbxMinMax.Checked then
      Options := Options + [voMinMax]
    else
      Options := Options - [voMinMax];
  Preview;
end;

{ - - - - - - - - - - - }

procedure TfrmValueOptions.edtMinExit(Sender: TObject);
begin
  with Item do
    try
      Min := StrToFloat(edtMin.Text);
    except
      edtMin.Text := Trim(Format(VFormat, [Min]));
    end;
  Preview;
end;

procedure TfrmValueOptions.edtMaxExit(Sender: TObject);
begin
  with Item do
    try
      Max := StrToFloat(edtMax.Text);
    except
      edtMax.Text := Trim(Format(VFormat, [Max]));
    end;
  Preview;
end;

procedure TfrmValueOptions.edtStepExit(Sender: TObject);
begin
  with Item do
    try
      Step := StrToFloat(edtStep.Text);
    except
      edtStep.Text := Trim(Format(VFormat, [Step]));
    end;
  Preview;
end;

{ - - - - - - - - - - - }

procedure TfrmValueOptions.btnBarLeftClick(Sender: TObject);
begin
  with cldColor, FData.FColor do
    begin
      Color := BarLeft;
      if Execute then
        BarLeft := Color;
    end;
  Preview;
end;

procedure TfrmValueOptions.btnBarRightClick(Sender: TObject);
begin
  with cldColor, FData.FColor do
    begin
      Color := BarRight;
      if Execute then
        BarRight := Color;
    end;
  Preview;
end;

procedure TfrmValueOptions.btnBarAvgClick(Sender: TObject);
begin
  with cldColor, FData.FColor do
    begin
      Color := BarAvg;
      if Execute then
        BarAvg := Color;
    end;
  Preview;
end;

procedure TfrmValueOptions.btnBarGridClick(Sender: TObject);
begin
  with cldColor, FData.FColor do
    begin
      Color := BarGrid;
      if Execute then
        BarGrid := Color;
    end;
  Preview;
end;

{ - - - - - - - - - - - }

procedure TfrmValueOptions.btnHistoryGraphClick(Sender: TObject);
begin
  with cldColor, FData.FColor do
    begin
      Color := HistoryGraph;
      if Execute then
        HistoryGraph := Color;
    end;
  Preview;
end;

procedure TfrmValueOptions.btnHistoryAverageClick(Sender: TObject);
begin
  with cldColor, FData.FColor do
    begin
      Color := HistoryAvg;
      if Execute then
        HistoryAvg := Color;
    end;
  Preview;
end;

procedure TfrmValueOptions.btnHistoryGridClick(Sender: TObject);
begin
  with cldColor, FData.FColor do
    begin
      Color := HistoryGrid;
      if Execute then
        HistoryGrid := Color;
    end;
  Preview;
end;

{ - - - - - - - - - - - }

procedure TfrmValueOptions.FormShow(Sender: TObject);
begin
  with TValueArea(pnlPreview.Controls[0]) do
    AddCellAt(0, Item.ID);
  with Item do
    begin
      gbxItem.Caption        := Format('[%s : %s]', [Caption, Hint]);
      cbxBar.Checked         := (voBar in Options);
      cbxBarGrid.Checked     := (voBarGrid in Options);
      cbxBarAvg.Checked      := (voBarAvg in Options);
      cbxHistory.Checked     := (voHistory in Options);
      cbxHistoryGrid.Checked := (voHistoryGrid in Options);
      cbxHistoryAvg.Checked  := (voHistoryAvg in Options);
      cbxHistoryLeft.Checked := (voHistoryLeft in Options);
      edtMin.Text            := Trim(Format(VFormat, [Min]));
      edtMax.Text            := Trim(Format(VFormat, [Max]));
      edtStep.Text           := Trim(Format(VFormat, [Step]));
      cbxMinMax.Checked      := (voMinMax in Options);
    end;
  cbxBarClick(Sender);
  cbxHistoryClick(Sender);
  with cldColor do
    begin
      CustomColors.Clear;
      CustomColors.Add(Format('ColorA=%x', [ColorToRGB(clBtnFace)]));
    end;
end;

procedure TfrmValueOptions.FormCreate(Sender: TObject);
begin
  FData := TValueArea(TfrmValueOptions(Sender).Owner).FData;
  with TValueArea.Create(pnlPreview) do
    begin
      FPreview := True;
      Align    := alClient;
      FData    := TfrmValueOptions(Self).FData;
      PopupMenu.Free;
    end;
end;

procedure TfrmValueOptions.Preview;
begin
  with TValueArea(pnlPreview.Controls[0]) do
    Update;
end;

procedure TfrmValueOptions.UpdateAppearance(oFont : TFont; oDialog, oWindow : TColor);
begin
  Font.Assign(oFont);
  Color := oDialog;
end;

procedure Register;
begin
  RegisterComponents('Samples', [TValueArea, TValueData]);
end;

end.
